#ifndef lint
static char sccsid[] = "@(#)tiff_write.c	1.11 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

/*
 * TIFF Library.
 *
 * Scanline-oriented Write Support
 */
#include "tiffio.h"

#define	WriteOK(fd, buf, size)	(write(fd, (char *)buf, size) == size)
#define	SeekOK(fd, off, whence)	(lseek(fd, (long)off, whence) == (long)off)
#define	howmany(x, y)	(((x)+((y)-1))/(y))
#define	roundup(x, y)	((((x)+((y)-1))/(y))*(y))

#define	STRIPINCR	20		/* expansion factor on strip array */

static	char module[] = "TIFFWriteScanline";
extern	char *malloc();

TIFFWriteScanline(tif, va_alist)
	register TIFF *tif;
	va_dcl
{
	register TIFFDirectory *td;
	int s, scanline, bitspersample, status = 0;
	struct strip *sp;
	va_list ap;

	if (tif->tif_mode == O_RDONLY) {
		TIFFError(module, "%s: File not open for writing",
		    tif->tif_name);
		return (-1);
	}
	td = tif->tif_dir;
	/*
	 * On the first write verify all the required information
	 * has been setup and initialize any data structures that
	 * had to wait until directory information was set.
	 * Note that a lot of our work is assumed to remain valid
	 * because we disallow any of the important parameters
	 * from changing after we start writing (i.e. once
	 * TIFF_BEENWRITING is set, TIFFSetField will only allow
	 * the image's length to be changed).
	 */
	if ((tif->tif_flags & TIFF_BEENWRITING) == 0) {
		if (!TIFFFieldSet(td, FIELD_IMAGEDIMENSIONS)) {
			TIFFError(module,
			    "%s: Must set \"ImageWidth\" before writing data",
			    tif->tif_name);
			return (-1);
		}
		if (!TIFFFieldSet(td, FIELD_PLANARCONFIG)) {
			TIFFError(module,
		    "%s: Must set \"PlanarConfiguration\" before writing data",
			    tif->tif_name);
			return (-1);
		}
		if (td->td_strips == 0) {
			td->td_nstrips = td->td_rowsperstrip == (u_long)-1 ||
			    td->td_imagelength == 0 ? 1 :
			    howmany(td->td_imagelength, td->td_rowsperstrip);
			if (td->td_planarconfig == PLANARCONFIG_SEPARATE)
				td->td_nstrips *= td->td_samplesperpixel;
			td->td_strips = (struct strip *)
			    malloc(td->td_nstrips * sizeof (struct strip));
			if (td->td_strips == NULL) {
				td->td_nstrips = 0;
				TIFFError(module,
				    "%s: No space for strip array",
				    tif->tif_name);
				return (-1);
			}
			/*
			 * Place data at the end-of-file
			 * (by setting offsets to zero).
			 */
			bzero((char *)td->td_strips,
			    td->td_nstrips * sizeof (struct strip));
			TIFFSetFieldBit(td, FIELD_STRIPOFFSETS);
			TIFFSetFieldBit(td, FIELD_STRIPBYTECOUNTS);
		}
		tif->tif_flags |= TIFF_BEENWRITING;
	}
	if ((tif->tif_flags & TIFF_BUFFERSETUP) == 0) {
		bitspersample = td->td_bitspersample;
		tif->tif_bytespersample =
		    bitspersample <= 8 ? 1 : bitspersample <= 16 ? 2 : 4;
		scanline = bitspersample * td->td_imagewidth;
		if (td->td_planarconfig == PLANARCONFIG_CONTIG)
			scanline *= td->td_samplesperpixel;
		scanline = howmany(scanline, 8);
		tif->tif_scanlinesize = scanline;
		/*
		 * Make raw data buffer at least 8K
		 */
		if (scanline < 8*1024)
			scanline = 8*1024;
		tif->tif_rawdata = malloc(scanline);
		if (tif->tif_rawdata == NULL) {
			TIFFError(module, "%s: No space for output buffer",
			    tif->tif_name);
			return (-1);
		}
		tif->tif_rawdatasize = scanline;
		tif->tif_rawcc = 0;
		tif->tif_rawcp = tif->tif_rawdata;
		tif->tif_datamask = (1 << td->td_bitspersample) - 1;
		tif->tif_flags |= TIFF_BUFFERSETUP;
	}
	/*
	 * If the buffer has data and this is
	 * the end of a strip, then flush.
	 */
	if (tif->tif_rawcc > 0 &&
	    ((tif->tif_row+1) % td->td_rowsperstrip) == 0) {
		if (tif->tif_encodestrip)
			(*tif->tif_encodestrip)(tif);
		if (!TIFFFlushData(tif))
			return (-1);
	}
	if (++tif->tif_row > td->td_imagelength)	/* extend image */
		td->td_imagelength = tif->tif_row;
	/*
	 * Check strip array to make sure there's space
	 * for data.  Note that when writing data with
	 * data in separate planes, we build the strip
	 * array as a packed array of strips -- the
	 * directory writing routine knows to rearrange
	 * the array before it's written out.
	 */
	s = tif->tif_row / td->td_rowsperstrip;
	if (td->td_planarconfig == PLANARCONFIG_SEPARATE)
		s += td->td_samplesperpixel;
	if (s > td->td_nstrips) {	/* grow strips array */
		s = STRIPINCR;
		if (td->td_planarconfig == PLANARCONFIG_SEPARATE)
			s += (td->td_samplesperpixel-1)*STRIPINCR;
		td->td_strips = (struct strip *)realloc(td->td_strips,
		    (td->td_nstrips + s) * sizeof (struct strip));
		if (td->td_strips == NULL) {
			td->td_nstrips = 0;
			TIFFError(module, "%s: No space to expand strip array",
			    tif->tif_name);
			return (-1);
		}
		bzero(td->td_strips + td->td_nstrips, s * sizeof (*sp));
		td->td_nstrips += s;
	}
	tif->tif_curstrip = &td->td_strips[tif->tif_row / td->td_rowsperstrip];
	va_start(ap);
	switch (td->td_planarconfig) {
	case PLANARCONFIG_CONTIG:
		status = ((*tif->tif_encoderow)(tif, va_arg(ap, u_char *)));
		break;
	case PLANARCONFIG_SEPARATE:
		for (s = 0; s < td->td_samplesperpixel; s++) {
			status = (*tif->tif_encoderow)(tif,
			    va_arg(ap, u_char *));
			if (status == 0)
				break;
			tif->tif_curstrip++;
		}
		break;
	}
	va_end(ap);
	return (status);
}

/*
 * Flush raw data to the file.
 */
TIFFFlushData(tif)
	TIFF *tif;
{
	struct strip *sp;
	TIFFDirectory *td = tif->tif_dir;

	sp = tif->tif_curstrip;
	if (sp->offset == 0 || tif->tif_curoff == 0) {
		/*
		 * No current offset, set the current strip.
		 */
		if (sp->offset != 0) {
			if (!SeekOK(tif->tif_fd, sp->offset, L_SET)) {
				TIFFError(module,
				    "%s: Seek error at scanline %d",
				    tif->tif_name, tif->tif_row);
				return (0);
			}
		} else
			sp->offset = lseek(tif->tif_fd, 0L, L_XTND);
		tif->tif_curoff = sp->offset;
	}
	if (!WriteOK(tif->tif_fd, tif->tif_rawdata, tif->tif_rawcc)) {
		TIFFError(module, "%s: Write error at scanline %d",
		    tif->tif_name, tif->tif_row);
		return (0);
	}
	tif->tif_curoff += tif->tif_rawcc;
	sp->bytecount += tif->tif_rawcc;
	tif->tif_rawcc = 0;
	tif->tif_rawcp = tif->tif_rawdata;
	return (1);
}
